from copy import deepcopy
from typing import List, Union

from metrics_layer.core.exceptions import (
    AccessDeniedOrDoesNotExistException,
    JoinError,
    QueryError,
)
from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.model.project import Project
from metrics_layer.core.sql.query_arbitrary_merged_queries import (
    MetricsLayerMergedQueries,
)
from metrics_layer.core.sql.query_base import QueryKindTypes
from metrics_layer.core.sql.resolve import SQLQueryResolver
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver


class ArbitraryMergedQueryResolver(SingleSQLQueryResolver):
    def __init__(
        self,
        merged_queries: List[dict],
        where: List[dict] = [],
        having: List[dict] = [],
        order_by: List[dict] = [],
        project: Union[Project, None] = None,
        connections: List = [],
        **kwargs,
    ):
        merged_queries = self.validate_arbitrary_merged_queries(merged_queries)
        self.merged_queries = merged_queries
        self.verbose = kwargs.get("verbose", False)
        self.where = where
        self.having = having
        self.order_by = order_by
        self.limit = kwargs.pop("limit", None)
        self.project = project
        self.connections = connections
        self.connection = None
        self.model = None
        # All queries are merged queries (obviously)
        self.query_kind = QueryKindTypes.merged
        self.kwargs = kwargs
        self._mapping_lookup = {}

    @property
    def mapping_lookup(self):
        self.get_query()
        return self._mapping_lookup

    def get_query(self, semicolon: bool = True):
        sub_queries = []
        source_fields = [jf["source_field"] for mq in self.merged_queries for jf in mq.get("join_fields", [])]
        primary_resolver = self._init_resolver(self.merged_queries[0], extra_lookup_dims=source_fields)
        self._add_mapping_lookup(primary_resolver)

        for i, merged_query in enumerate(self.merged_queries):
            extra_lookup_dims = [jf["field"] for jf in merged_query.get("join_fields", [])]
            resolver = self._init_resolver(merged_query, extra_lookup_dims=extra_lookup_dims)
            self._add_mapping_lookup(resolver)

            self.connection = resolver.connection
            join_fields = merged_query.get("join_fields", [])
            if i > 0:
                join_fields = self._resolve_join_fields_mappings(primary_resolver, resolver, join_fields, i)

            sub_query = resolver.get_query(semicolon=False)
            sub_queries.append(
                {
                    "metrics": resolver.metrics,
                    "dimensions": resolver.dimensions,
                    "cte_alias": f"merged_query_{i}",
                    "query": sub_query,
                    "join_fields": join_fields,
                    "join_type": merged_query.get("join_type"),
                }
            )

        mapping_lookup = self._mapping_lookup
        self.model = resolver.model
        clean_where = [{**w, "field": mapping_lookup.get(w["field"].lower(), w["field"])} for w in self.where]
        clean_having = [
            {**h, "field": mapping_lookup.get(h["field"].lower(), h["field"])} for h in self.having
        ]
        clean_order_by = [
            {**o, "field": mapping_lookup.get(o["field"].lower(), o["field"])} for o in self.order_by
        ]
        merged_queries_resolver = MetricsLayerMergedQueries(
            {
                "merged_queries": sub_queries,
                "query_type": resolver.query_type,
                "where": clean_where,
                "having": clean_having,
                "order_by": clean_order_by,
                "limit": self.limit,
                "project": self.project,
            }
        )
        # Druid does not allow semicolons
        if resolver.query_type in Definitions.no_semicolon_warehouses:
            semicolon = False

        query = merged_queries_resolver.get_query(semicolon=semicolon)
        return query

    def _resolve_join_fields_mappings(
        self,
        primary_resolver: SQLQueryResolver,
        secondary_resolver: SQLQueryResolver,
        join_fields: List[dict],
        query_number: int,
    ):
        join_fields_resolved = []
        for join_field in join_fields:
            self._resolve_join_logic(secondary_resolver, "field", join_field, query_number + 1)
            self._resolve_join_logic(primary_resolver, "source_field", join_field, 1)
            join_fields_resolved.append(join_field)
        return join_fields_resolved

    # Note: this mutates join_field
    def _resolve_join_logic(self, resolver, key, join_field, query_number: int):
        if mapping_object := resolver.mapping_lookup.get(join_field[key]):
            try:
                join_field[key] = next((f for f in mapping_object["fields"] if f in resolver.dimensions))
            except StopIteration:
                self._raise_join_error(join_field[key], query_number)
        else:
            try:
                field = self.project.get_field(join_field[key])
                dimension_ids = [
                    resolver.field_object_lookup[d].id() if d in resolver.field_object_lookup else d
                    for d in resolver.dimensions
                ]
                if field.id() not in dimension_ids:
                    self._raise_join_error(join_field[key], query_number)
            except AccessDeniedOrDoesNotExistException:
                self._raise_join_error(join_field[key], query_number)

    def _raise_join_error(self, field_name: str, query_number):
        raise JoinError(
            f"Join field {field_name} not found in the query number {query_number}. To be used as a join the"
            f" field must be included in query {query_number}."
        )

    def _init_resolver(self, merged_query: dict, extra_lookup_dims: list = []):
        kws = {
            **self.kwargs,
            "limit": merged_query.get("limit"),
            "mapping_lookup_dimensions": extra_lookup_dims,
            "return_pypika_query": True,
        }
        return SQLQueryResolver(
            metrics=merged_query.get("metrics", []),
            dimensions=merged_query.get("dimensions", []),
            where=merged_query.get("where", []),
            having=merged_query.get("having", []),
            order_by=merged_query.get("order_by", []),
            project=self.project,
            connections=self.connections,
            **kws,
        )

    def _add_mapping_lookup(self, resolver: SQLQueryResolver):
        where_fields = [f["field"] for f in resolver.where]
        order_by_fields = [f["field"] for f in resolver.order_by]
        for mapping_name, mapping_field in resolver.mapping_lookup.items():
            if mapping_name not in self._mapping_lookup:
                if any(d in mapping_field["fields"] for d in resolver.dimensions):
                    field_name = next((f for f in mapping_field["fields"] if f in resolver.dimensions))
                elif any(f in mapping_field["fields"] for f in where_fields):
                    field_name = next((f for f in mapping_field["fields"] if f in where_fields))
                elif any(f in mapping_field["fields"] for f in order_by_fields):
                    field_name = next((f for f in mapping_field["fields"] if f in order_by_fields))
                self._mapping_lookup[mapping_name] = field_name

    @staticmethod
    def validate_arbitrary_merged_queries(merged_queries: list):
        cleaned_merged_queries = []
        for i, merged_query in enumerate(deepcopy(merged_queries)):
            if not isinstance(merged_query, dict):
                raise QueryError(
                    f"merged_queries must be a list of dictionaries. Item {i} is not a dictionary."
                )
            if not all(
                key in merged_query and isinstance(merged_query[key], list)
                for key in ["metrics", "dimensions"]
            ):
                raise QueryError(f"Each item in merged_queries must have 'metrics' and 'dimensions' keys.")
            if merged_query.get("funnel"):
                raise QueryError(
                    f"Each item in merged_queries must not have 'funnel' key. Funnels are not supported in"
                    f" merged queries."
                )
            if not isinstance(merged_query.get("join_fields", []), list):
                raise QueryError(f"Each item in merged_queries must have 'join_fields' key as a list.")
            if i > 0 and not merged_query.get("join_fields", []):
                raise QueryError(f"Each item in merged_queries after the first must have 'join_fields' key.")
            if i > 0:
                join_fields = []
                for join_field in merged_query.get("join_fields", []):
                    if ("field" not in join_field or not isinstance(join_field.get("field"), str)) or (
                        "source_field" not in join_field
                        or not isinstance(join_field.get("source_field"), str)
                    ):
                        raise QueryError(
                            f"Each item in merged_queries after the first must have the 'join_fields' key"
                            f" with the keys 'field' and 'source_field', both of which are field ids"
                            f" (strings)"
                        )
                    join_fields.append(
                        {
                            "field": join_field["field"].lower(),
                            "source_field": join_field["source_field"].lower(),
                        }
                    )
                merged_query["join_fields"] = join_fields
            cleaned_merged_queries.append(merged_query)
        return cleaned_merged_queries



import re
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Callable

import pandas as pd
import pendulum
from pypika import Criterion
from pypika.functions import Lower
from pypika.terms import LiteralValue

from metrics_layer.core.exceptions import QueryError

from .base import MetricsLayerBase
from .week_start_day_types import WeekStartDayTypes

if TYPE_CHECKING:
    from metrics_layer.core.model.view import View


class LiteralValueCriterion(Criterion):
    def __init__(self, sql_query: str, alias: str = None) -> None:
        """A wrapper for a literal value criterion which is a string of valid sql"""
        super().__init__(alias)
        self.sql_query = sql_query

    def get_sql(self, **kwargs):
        return self.sql_query


class MetricsLayerFilterGroupLogicalOperatorType:
    and_ = "AND"
    or_ = "OR"
    options = [and_, or_]


class MetricsLayerFilterExpressionType(str, Enum):
    Unknown = "UNKNOWN"
    LessThan = "less_than"
    LessOrEqualThan = "less_or_equal_than"
    EqualTo = "equal_to"
    NotEqualTo = "not_equal_to"
    GreaterOrEqualThan = "greater_or_equal_than"
    GreaterThan = "greater_than"
    Like = "like"
    Contains = "contains"
    DoesNotContain = "does_not_contain"
    ContainsCaseInsensitive = "contains_case_insensitive"
    DoesNotContainCaseInsensitive = "does_not_contain_case_insensitive"
    StartsWith = "starts_with"
    EndsWith = "ends_with"
    DoesNotStartWith = "does_not_start_with"
    DoesNotEndWith = "does_not_end_with"
    StartsWithCaseInsensitive = "starts_with_case_insensitive"
    EndsWithCaseInsensitive = "ends_with_case_insensitive"
    DoesNotStartWithCaseInsensitive = "does_not_start_with_case_insensitive"
    DoesNotEndWithCaseInsensitive = "does_not_end_with_case_insensitive"
    IsNull = "is_null"
    IsNotNull = "is_not_null"
    IsIn = "isin"
    IsNotIn = "isnotin"
    IsInQuery = "is_in_query"
    IsNotInQuery = "is_not_in_query"
    BooleanTrue = "boolean_true"
    BooleanFalse = "boolean_false"
    IsTrue = "is_true"
    IsFalse = "is_false"
    Matches = "matches"

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value.lower() == other.value.lower()

    @classmethod
    def parse(cls, value: str):
        try:
            return next(e for e in cls if e.value.lower() == value)
        except StopIteration:
            return cls.Unknown


class FilterInterval(str, Enum):
    unknown = "UNKNOWN"
    day = "day"
    week = "week"
    month = "month"
    quarter = "quarter"
    year = "year"

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value.lower() == other.value.lower()

    @staticmethod
    def plural(value: str):
        return value.lower() if value[-1] == "s" else f"{value.lower()}s"

    @staticmethod
    def singular(value: str):
        return value.lower()[:-1] if value[-1] == "s" else value.lower()

    @classmethod
    def all(cls):
        return [i for e in cls for i in [e.value.lower(), f"{e.value.lower()}s"]]

    @classmethod
    def parse(cls, value: str):
        try:
            return next(e for e in cls if value in {e.value.lower(), f"{e.value.lower()}s"})
        except StopIteration:
            return cls.unknown


class Filter(MetricsLayerBase):
    week_start_day_default = pendulum.MONDAY
    week_end_day_default = pendulum.SUNDAY
    week_start_day_lookup = {
        WeekStartDayTypes.monday: pendulum.MONDAY,
        WeekStartDayTypes.tuesday: pendulum.TUESDAY,
        WeekStartDayTypes.wednesday: pendulum.WEDNESDAY,
        WeekStartDayTypes.thursday: pendulum.THURSDAY,
        WeekStartDayTypes.friday: pendulum.FRIDAY,
        WeekStartDayTypes.saturday: pendulum.SATURDAY,
        WeekStartDayTypes.sunday: pendulum.SUNDAY,
    }

    def __init__(self, definition: dict = {}) -> None:
        self.validate(definition)
        super().__init__(definition)

    def validate(self, definition: dict):
        required_keys = ["field", "value"]
        for k in required_keys:
            if k not in definition:
                raise QueryError(f"Filter missing required key '{k}' The filter passed was {definition}")

    def filter_dict(self, json_safe: bool = False) -> list:
        filter_dict = self._filter_dict(self.field, self.value, self.week_start_day, self.timezone)
        if isinstance(filter_dict, dict):
            filter_list = [filter_dict]
        else:
            filter_list = filter_dict
        return [self._clean_filter_dict({**self._definition, **f}, json_safe) for f in filter_list]

    @staticmethod
    def _clean_filter_dict(filter_dict: dict, json_safe: bool):
        if json_safe and "expression" in filter_dict:
            filter_dict["expression"] = filter_dict["expression"].value
        return filter_dict

    @staticmethod
    def _set_week_start_day(week_start_day: str):
        week_start_day = Filter.week_start_day_lookup.get(week_start_day, Filter.week_start_day_default)
        pendulum.week_starts_at(week_start_day)
        week_end_day = week_start_day - 1 if week_start_day != 0 else 6
        pendulum.week_ends_at(week_end_day)

    @staticmethod
    def _reset_week_start_day():
        pendulum.week_starts_at(Filter.week_start_day_default)
        pendulum.week_ends_at(Filter.week_end_day_default)

    @staticmethod
    def _end_date(lag: int, date_part: str, tz: str):
        plural_date_part = FilterInterval.plural(date_part)
        singular_date_part = FilterInterval.singular(date_part)
        now = Filter._today(tz=tz)

        if singular_date_part == "quarter":
            date = now.subtract(months=3 * lag).last_of("quarter")
        else:
            date = now.subtract(**{plural_date_part: lag}).end_of(singular_date_part)
        return Filter._date_to_string(date)

    @staticmethod
    def _start_date(lag: int, date_part: str, tz: str, return_date=False):
        plural_date_part = FilterInterval.plural(date_part)
        singular_date_part = FilterInterval.singular(date_part)
        now = Filter._today(tz=tz)

        if singular_date_part == "quarter":
            date = now.subtract(months=3 * lag).first_of("quarter")
        else:
            date = now.subtract(**{plural_date_part: lag}).start_of(singular_date_part)
        if return_date:
            return date
        return Filter._date_to_string(date)

    @staticmethod
    def _add_to_end_date(date, lag: int, date_part: str):
        plural_date_part = FilterInterval.plural(date_part)
        singular_date_part = FilterInterval.singular(date_part)
        if singular_date_part == "quarter":
            date = date.add(months=3 * lag).last_of("quarter")
        else:
            date = date.add(**{plural_date_part: lag}).end_of(singular_date_part)
        return date

    @staticmethod
    def parse_date_condition(date_condition: str, tz: str) -> list:
        if tz is None:
            tz = "UTC"

        if " until " in str(date_condition):
            first, second = date_condition.split(" until ")

            start_expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
            parsed_first = Filter.parse_date_condition(first, tz)
            if parsed_first is None:
                start_value = Filter._parse_date_string(first)
                first_filter = (start_expression, start_value)
            else:
                first_filter = parsed_first[0]

            end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
            parsed_second = Filter.parse_date_condition(second, tz)
            if parsed_second is None:
                end_value = Filter._parse_date_string(second)
                second_filter = (end_expression, end_value)
            else:
                second_filter = parsed_second[-1]

            return [first_filter, second_filter]

        if date_condition == "today":
            expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
            cleaned_value = Filter._start_date(lag=0, date_part=FilterInterval.day, tz=tz)
            end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
            end_value = Filter._end_date(lag=0, date_part=FilterInterval.day, tz=tz)
            return [(expression, cleaned_value), (end_expression, end_value)]

        if date_condition == "yesterday":
            start_expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
            start_value = Filter._start_date(lag=1, date_part=FilterInterval.day, tz=tz)

            end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
            end_value = Filter._end_date(lag=1, date_part=FilterInterval.day, tz=tz)
            return [(start_expression, start_value), (end_expression, end_value)]

        # Match regex patterns to the various date filters and when patterns
        # match assume it's a date not a string comparison
        interval_condition = "|".join(FilterInterval.all())
        interval_ago_for_result = Filter._parse_n_interval_ago_for(date_condition, interval_condition, tz=tz)
        if interval_ago_for_result:
            return interval_ago_for_result

        interval_modifier_result = Filter._parse_n_interval_modifier(
            date_condition, interval_condition, tz=tz
        )
        if interval_modifier_result:
            return interval_modifier_result

        n_interval_result = Filter._parse_n_interval(date_condition, interval_condition, tz=tz)
        if n_interval_result:
            return n_interval_result

    @staticmethod
    def _parse_n_interval(date_condition: str, interval_condition: str, tz: str):
        regex = rf"(\d+|this|last)\s+({interval_condition})"  # noqa
        result = re.search(regex, str(date_condition))
        if not result:
            return

        n = result.group(1)
        date_part = result.group(2)

        if n == "this":
            lag = 0
        elif n == "last":
            lag = 1
        else:
            lag = int(n) - 1

        start_expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
        start_value = Filter._start_date(lag=lag, date_part=date_part, tz=tz)
        result = [(start_expression, start_value)]

        end_lag = lag if n == "last" else 0
        end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
        end_value = Filter._end_date(lag=end_lag, date_part=date_part, tz=tz)
        result.append((end_expression, end_value))

        return result

    @staticmethod
    def _parse_n_interval_modifier(date_condition: str, interval_condition: str, tz: str):
        regex = rf"(\d+\s+|this\s+|last\s+|)({interval_condition})\s+(ago\s+to\s+date|ago|to\s+date)"  # noqa
        result = re.search(regex, str(date_condition))
        if not result:
            return

        n = result.group(1).strip()
        date_part = result.group(2)
        modifier = result.group(3)

        if n in {"this", ""}:
            lag = 0
        elif n == "last":
            lag = 1
        elif modifier == "to date" and n != "last":
            lag = int(n) - 1
        else:
            lag = int(n)

        start_expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
        start_date = Filter._start_date(lag=lag, date_part=date_part, tz=tz, return_date=True)
        result = [(start_expression, Filter._date_to_string(start_date))]

        if modifier == "ago to date" or modifier == "to date":
            end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan

            # We need to figure out how many days are between now and the start of the period
            start_of_period = Filter._start_date(lag=0, date_part=date_part, tz=tz, return_date=True)
            time_change = Filter._today(tz=tz) - start_of_period
            lag = time_change.days - 1 if time_change.days > 0 else 0
            end_date = Filter._add_to_end_date(start_date, lag=lag, date_part="day")
            result.append((end_expression, Filter._date_to_string(end_date)))

        elif modifier == "ago":
            end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
            end_value = Filter._end_date(lag=lag, date_part=date_part, tz=tz)
            result.append((end_expression, end_value))

        return result

    @staticmethod
    def _parse_n_interval_ago_for(date_condition: str, interval_condition: str, tz: str):
        regex = rf"(\d+)\s+({interval_condition})\s+ago\s+for\s+(\d+)\s+({interval_condition})"  # noqa
        result = re.search(regex, str(date_condition))
        if not result:
            return

        first_n = int(result.group(1))
        first_date_part = result.group(2)
        second_n = int(result.group(3)) - 1
        second_date_part = result.group(4)

        start_expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
        start_date = Filter._start_date(lag=first_n, date_part=first_date_part, tz=tz, return_date=True)
        result = [(start_expression, Filter._date_to_string(start_date))]

        end_expression = MetricsLayerFilterExpressionType.LessOrEqualThan
        end_date = Filter._add_to_end_date(start_date, lag=second_n, date_part=second_date_part)
        result.append((end_expression, Filter._date_to_string(end_date)))

        return result

    @staticmethod
    def _filter_dict(field: str, value: str, week_start_day: str = None, tz: str = None):
        # TODO more advanced parsing similar to
        # ref: https://docs.looker.com/reference/field-params/filters

        # If the value is an empty dict, we should return None because the filters will be skipped
        if str(value) == "":
            return {}

        _symbol_to_filter_type_lookup = {
            "<=": MetricsLayerFilterExpressionType.LessOrEqualThan,
            ">=": MetricsLayerFilterExpressionType.GreaterOrEqualThan,
            "<>": MetricsLayerFilterExpressionType.NotEqualTo,
            "!=": MetricsLayerFilterExpressionType.NotEqualTo,
            "-": MetricsLayerFilterExpressionType.NotEqualTo,
            "=": MetricsLayerFilterExpressionType.EqualTo,
            ">": MetricsLayerFilterExpressionType.GreaterThan,
            "<": MetricsLayerFilterExpressionType.LessThan,
        }
        Filter._set_week_start_day(week_start_day)
        date_condition = Filter.parse_date_condition(value, tz=tz)
        Filter._reset_week_start_day()

        first_word = str(value).split(" ")[0]
        first_two_words = " ".join(str(value).split(" ")[:2])

        split_expression = ","
        # Handle field to field comparison
        if isinstance(value, LiteralValue):
            expression = MetricsLayerFilterExpressionType.EqualTo
            cleaned_value = value

        # Handle null conditional
        elif value == "NULL" or value is None:
            expression = MetricsLayerFilterExpressionType.IsNull
            cleaned_value = None

        elif value == "-NULL":
            expression = MetricsLayerFilterExpressionType.IsNotNull
            cleaned_value = None

        # Handle boolean True and False
        elif value in {True, False, "TRUE", "FALSE"}:  # noqa
            expression = MetricsLayerFilterExpressionType.EqualTo
            if value in {True, "TRUE"}:
                cleaned_value = True
            elif value in {False, "FALSE"}:
                cleaned_value = False

        # Handle date conditions
        elif date_condition:
            multiple_filter_dicts = []
            for expression, cleaned_value in date_condition:
                filter_dict = {"field": field, "expression": expression, "value": cleaned_value}
                multiple_filter_dicts.append(filter_dict)
            if len(multiple_filter_dicts) == 1:
                return multiple_filter_dicts[0]
            return multiple_filter_dicts

        # Handle date after and before
        elif first_word in {"after", "before", "on"} or first_two_words in {"not on"}:
            cleaned_value = Filter._parse_date_string(value.split(" ")[-1])
            if first_word == "after":
                expression = MetricsLayerFilterExpressionType.GreaterOrEqualThan
            elif first_word == "before":
                expression = MetricsLayerFilterExpressionType.LessOrEqualThan
            elif first_word == "on":
                expression = MetricsLayerFilterExpressionType.EqualTo
            else:
                expression = MetricsLayerFilterExpressionType.NotEqualTo

        # isin for strings
        elif len(value.split(split_expression)) > 1:
            if all(category.strip()[0] == "-" for category in value.split(split_expression)):
                expression = MetricsLayerFilterExpressionType.IsNotIn
                cleaned_value = [
                    f"{category.strip()[1:].strip()}" for category in value.split(split_expression)
                ]

            elif any(category.strip()[0] == "-" for category in value.split(split_expression)):
                raise QueryError("Invalid filter some elements are negated with '-' and some are not")

            else:
                expression = MetricsLayerFilterExpressionType.IsIn
                cleaned_value = [f"{category.strip()}" for category in value.split(split_expression)]

        # Numeric parsing for less than or equal to, greater than or equal to, not equal to
        elif value[:2] in {"<=", ">=", "<>", "!="}:
            expression = _symbol_to_filter_type_lookup[value[:2]]
            cleaned_value = pd.to_numeric(value[2:])

        # Numeric parsing for equal to, less than, greater than
        elif value[0] in {"=", ">", "<"}:
            expression = _symbol_to_filter_type_lookup[value[0]]
            cleaned_value = pd.to_numeric(value[1:])

        # String parsing for NOT equal to
        elif value[0] == "-":
            if value[1] == "%" and value[-1] == "%":
                expression = MetricsLayerFilterExpressionType.DoesNotContainCaseInsensitive
                cleaned_value = value[2:-1]
            elif value[1] == "%":
                expression = MetricsLayerFilterExpressionType.DoesNotEndWithCaseInsensitive
                cleaned_value = value[2:]
            elif value[-1] == "%":
                expression = MetricsLayerFilterExpressionType.DoesNotStartWithCaseInsensitive
                cleaned_value = value[1:-1]
            else:
                expression = _symbol_to_filter_type_lookup[value[0]]
                cleaned_value = value[1:]

        else:
            if value[0] == "%" and value[-1] == "%":
                expression = MetricsLayerFilterExpressionType.ContainsCaseInsensitive
                cleaned_value = value[1:-1]
            elif value[0] == "%":
                expression = MetricsLayerFilterExpressionType.EndsWithCaseInsensitive
                cleaned_value = value[1:]
            elif value[-1] == "%":
                expression = MetricsLayerFilterExpressionType.StartsWithCaseInsensitive
                cleaned_value = value[:-1]
            else:
                expression = MetricsLayerFilterExpressionType.EqualTo
                cleaned_value = value

        return {"field": field, "expression": expression, "value": cleaned_value}

    @staticmethod
    def _parse_date_string(date_string: str):
        parsed_date = datetime.strptime(date_string, "%Y-%m-%d")
        return Filter._date_to_string(parsed_date)

    @staticmethod
    def _today(tz):
        timezone = pendulum.timezone(tz)
        return pendulum.now(timezone)

    @staticmethod
    def _date_to_string(date_obj):
        return date_obj.strftime("%Y-%m-%dT%H:%M:%S")

    @staticmethod
    def translate_looker_filters_to_sql(
        sql: str,
        filters: list,
        view: "View",
        else_0: bool = False,
        sql_replacement_func: Callable = lambda x: x,
    ):
        case_sql = "case when "
        conditions = []
        for f in filters:
            # All filters must have both of these keys, otherwise they are invalid
            if not all(k in f for k in ["field", "value"]):
                continue

            if "." not in f["field"]:
                field_id = f'{view.name}.{f["field"]}'
            else:
                field_id = f["field"]
            try:
                field = view.project.get_field(field_id)
                field_datatype = field.type
            except Exception:
                field_datatype = "unknown"
            value = sql_replacement_func(f["value"])
            filter_dict = Filter._filter_dict(f["field"], value, f.get("week_start_day"), f.get("timezone"))
            if isinstance(filter_dict, dict):
                filter_list = [filter_dict]
            else:
                filter_list = filter_dict

            for filter_obj in filter_list:
                if filter_obj != {}:
                    field_reference = "${" + f["field"] + "}"
                    condition_value = Filter.sql_query(
                        field_reference, filter_obj["expression"], filter_obj["value"], field_datatype
                    )
                    condition = f"{condition_value}"
                    conditions.append(condition)

        # Add the filter conditions AND'd together
        case_sql += " and ".join(conditions)
        # Add the result from the sql arg + imply NULL for anything not hitting the filter condition
        if else_0:
            case_sql += f" then {sql} else 0 end"
        else:
            case_sql += f" then {sql} end"

        return case_sql

    @staticmethod
    def sql_query(sql_to_compare: str, expression_type: str, value, field_datatype: str):
        field = LiteralValue(sql_to_compare)
        if (
            expression_type
            in {MetricsLayerFilterExpressionType.IsIn, MetricsLayerFilterExpressionType.IsNotIn}
            and field_datatype == "number"
        ):
            value = [pd.to_numeric(v) for v in value]
        criterion_strategies = {
            MetricsLayerFilterExpressionType.LessThan: lambda f: f < value,
            MetricsLayerFilterExpressionType.LessOrEqualThan: lambda f: f <= value,
            MetricsLayerFilterExpressionType.EqualTo: lambda f: f == value,
            MetricsLayerFilterExpressionType.NotEqualTo: lambda f: f != value,
            MetricsLayerFilterExpressionType.GreaterOrEqualThan: lambda f: f >= value,
            MetricsLayerFilterExpressionType.GreaterThan: lambda f: f > value,
            MetricsLayerFilterExpressionType.Like: lambda f: f.like(value),
            MetricsLayerFilterExpressionType.Contains: lambda f: f.like(f"%{value}%")
            if isinstance(value, str)
            else f.like(value),
            MetricsLayerFilterExpressionType.DoesNotContain: lambda f: f.not_like(f"%{value}%")
            if isinstance(value, str)
            else f.not_like(value),
            MetricsLayerFilterExpressionType.ContainsCaseInsensitive: lambda f: Lower(f).like(
                Lower(f"%{value}%") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.DoesNotContainCaseInsensitive: lambda f: Lower(f).not_like(
                Lower(f"%{value}%") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.StartsWith: lambda f: f.like(f"{value}%")
            if isinstance(value, str)
            else f.like(value),
            MetricsLayerFilterExpressionType.EndsWith: lambda f: f.like(f"%{value}")
            if isinstance(value, str)
            else f.like(value),
            MetricsLayerFilterExpressionType.DoesNotStartWith: lambda f: f.not_like(f"{value}%")
            if isinstance(value, str)
            else f.not_like(value),
            MetricsLayerFilterExpressionType.DoesNotEndWith: lambda f: f.not_like(f"%{value}")
            if isinstance(value, str)
            else f.not_like(value),
            MetricsLayerFilterExpressionType.StartsWithCaseInsensitive: lambda f: Lower(f).like(
                Lower(f"{value}%") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.EndsWithCaseInsensitive: lambda f: Lower(f).like(
                Lower(f"%{value}") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.DoesNotStartWithCaseInsensitive: lambda f: Lower(f).not_like(
                Lower(f"{value}%") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.DoesNotEndWithCaseInsensitive: lambda f: Lower(f).not_like(
                Lower(f"%{value}") if isinstance(value, str) else Lower(value)
            ),
            MetricsLayerFilterExpressionType.IsNull: lambda f: f.isnull(),
            MetricsLayerFilterExpressionType.IsNotNull: lambda f: f.notnull(),
            MetricsLayerFilterExpressionType.IsIn: lambda f: f.isin(value),
            MetricsLayerFilterExpressionType.IsNotIn: lambda f: f.isin(value).negate(),
            MetricsLayerFilterExpressionType.BooleanTrue: lambda f: LiteralValueCriterion(f),
            MetricsLayerFilterExpressionType.BooleanFalse: lambda f: f.negate(),
            MetricsLayerFilterExpressionType.IsTrue: lambda f: LiteralValueCriterion(f),
            MetricsLayerFilterExpressionType.IsFalse: lambda f: f.negate(),
        }

        try:
            return criterion_strategies[expression_type](field)
        except KeyError:
            raise QueryError(f"Unknown filter expression_type: {expression_type}.")



from collections import defaultdict

from pypika import AliasedQuery, Criterion, Order
from pypika.terms import LiteralValue

from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.model.filter import LiteralValueCriterion
from metrics_layer.core.sql.query_base import MetricsLayerQueryBase
from metrics_layer.core.sql.query_dialect import (
    NullSorting,
    if_null_lookup,
    query_lookup,
)


class MetricsLayerMergedResultsQuery(MetricsLayerQueryBase):
    """ """

    def __init__(self, definition: dict) -> None:
        self.query_lookup = query_lookup
        super().__init__(definition)

    def get_query(self, semicolon: bool = True):
        # Build the base_cte table from the referenced queries + join them with all dimensions
        base_cte_query = self.build_cte_from()

        # Add all columns in the SELECT clause
        select = self.get_select_columns()
        complete_query = base_cte_query.select(*select)

        if self.having:
            where = self.get_where_with_aliases(self.having, project=self.project)
            complete_query = complete_query.where(Criterion.all(where))

        if self.order_by:
            for arg in self.order_by:
                try:
                    field = self.project.get_field(arg["field"])
                    arg["field"] = field.alias(with_view=True)
                except Exception:
                    continue
                order = Order.desc if arg.get("sort") and arg.get("sort").lower() == "desc" else Order.asc
                complete_query = complete_query.orderby(
                    LiteralValue(arg["field"]), order=order, nulls=NullSorting.last
                )

        completed_query = complete_query.limit(self.limit)
        if self.return_pypika_query:
            return completed_query

        sql = str(completed_query)
        if semicolon:
            sql += ";"
        return sql

    def build_cte_from(self):
        base_cte_query = self._base_query()
        for join_hash, query in self.queries_to_join.items():
            base_cte_query = base_cte_query.with_(query, join_hash)

        for i, join_hash in enumerate(self.join_hashes):
            if i == 0:
                base_cte_query = base_cte_query.from_(AliasedQuery(join_hash))
            else:
                no_dimensions = all(len(v) == 0 for v in self.query_dimensions.values())
                # We have to do this because Redshift doesn't support a full outer join
                # of two CTE's without dimensions using 1=1
                if self.query_type == Definitions.redshift and no_dimensions:
                    base_cte_query = base_cte_query.join(AliasedQuery(join_hash)).cross()
                else:
                    criteria = self._build_join_criteria(self.join_hashes[0], join_hash, no_dimensions)
                    base_cte_query = base_cte_query.outer_join(AliasedQuery(join_hash)).on(criteria)

        return base_cte_query

    def _build_join_criteria(self, first_query_alias, second_query_alias, no_dimensions: bool):
        # No dimensions to join on, the query results must be just one number each
        if no_dimensions:
            return LiteralValueCriterion("1=1")

        join_criteria = []
        for i in range(len(self.query_dimensions[first_query_alias])):
            first_field = self.query_dimensions[first_query_alias][i]
            second_field = self.query_dimensions[second_query_alias][i]
            first_alias_and_id = f"{first_query_alias}.{first_field.alias(with_view=True)}"
            second_alias_and_id = f"{second_query_alias}.{second_field.alias(with_view=True)}"
            # We need to add casting for differing datatypes on dimension groups for BigQuery
            if Definitions.bigquery == self.query_type and first_field.datatype != second_field.datatype:
                join_logic = (
                    f"CAST({first_alias_and_id} AS TIMESTAMP)=CAST({second_alias_and_id} AS TIMESTAMP)"
                )
            else:
                join_logic = f"{first_alias_and_id}={second_alias_and_id}"
            join_criteria.append(join_logic)

        return LiteralValueCriterion(" and ".join(join_criteria))

    # Code to handle SELECT portion of query
    def get_select_columns(self):
        select = []
        existing_aliases = []
        for join_hash, field_set in sorted(self.query_metrics.items()):
            for field in field_set:
                alias = field.alias(with_view=True)
                if alias not in existing_aliases:
                    select.append(self.sql(f"{join_hash}.{alias}", alias=alias))
                    existing_aliases.append(alias)

        # Map the dimensions to their counterparts (if present) for the "if null" clauses
        mapping_lookup = defaultdict(list)
        for _, fields in self.query_dimensions.items():
            for field in fields:
                field_key = f"{field.view.name}.{field.name}"
                if field.dimension_group:
                    field_key_dim_group = f"{field_key}_{field.dimension_group}"
                else:
                    field_key_dim_group = field_key
                for mapped_field in self.mapping_lookup.get(field_key, []):
                    mapped_field_key = mapped_field["field"]
                    if field.dimension_group:
                        mapped_field_key += f"_{field.dimension_group}"
                    mapping_lookup[field_key_dim_group].append(
                        {"cte": mapped_field["cte"], "field": mapped_field_key}
                    )

        dimension_sql = {}
        all_dimension_ids = [field.id() for fields in self.query_dimensions.values() for field in fields]
        for join_hash, field_set in sorted(self.query_dimensions.items()):
            for field in field_set:
                alias = field.alias(with_view=True)
                if alias not in dimension_sql:
                    dimension_sql[alias] = f"{join_hash}.{alias}"
                else:
                    if_null_func = if_null_lookup[self.query_type]
                    dimension_sql[alias] = f"{if_null_func}({dimension_sql[alias]}, {join_hash}.{alias})"

                if field.id() in mapping_lookup:
                    present_fields = [
                        f for f in mapping_lookup[field.id()] if f["field"] in all_dimension_ids
                    ]
                    if_null_func = if_null_lookup[self.query_type]
                    nested_sql = self.nested_if_null(present_fields, if_null_func)
                    dimension_sql[alias] = f"{if_null_func}({join_hash}.{alias}, {nested_sql})"

        for alias, sql in dimension_sql.items():
            select.append(self.sql(sql, alias=alias))
            existing_aliases.append(alias)

        for field in self.merged_metrics:
            alias = field.alias(with_view=True)
            if alias not in existing_aliases:
                select.append(self.sql(field.strict_replaced_query(), alias=alias))
                existing_aliases.append(alias)

        return select

    @staticmethod
    def nested_if_null(aliases, if_null_func):
        first_alias = aliases[0]["cte"] + "." + aliases[0]["field"].replace(".", "_")
        if len(aliases) == 1:
            return first_alias
        else:
            return f"{if_null_func}({first_alias}, {MetricsLayerMergedResultsQuery.nested_if_null(aliases[1:], if_null_func)})"  # noqa



from pypika import AliasedQuery, Criterion, Order
from pypika.terms import LiteralValue

from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.model.filter import LiteralValueCriterion
from metrics_layer.core.model.join import ZenlyticJoinType
from metrics_layer.core.sql.query_base import MetricsLayerQueryBase
from metrics_layer.core.sql.query_dialect import NullSorting, query_lookup


class MetricsLayerMergedQueries(MetricsLayerQueryBase):
    """Resolve the SQL query for multiple, arbitrary merged queries"""

    def __init__(self, definition: dict) -> None:
        self.query_lookup = query_lookup
        super().__init__(definition)

    def get_query(self, semicolon: bool = True):
        # Build the base_cte table from the referenced queries + join them with all dimensions
        base_cte_query = self.build_cte_from()

        # Add all columns in the SELECT clause
        select = self.get_select_columns()
        complete_query = base_cte_query.select(*select)
        if self.where:
            where = self.get_where_with_aliases(
                self.where,
                project=self.project,
                cte_alias_lookup=self.cte_alias_lookup,
                raise_if_not_in_lookup=True,
            )
            complete_query = complete_query.where(Criterion.all(where))

        if self.having:
            # These become a where because we're only dealing with aliases, post aggregation
            having = self.get_where_with_aliases(
                self.having,
                project=self.project,
                cte_alias_lookup=self.cte_alias_lookup,
                raise_if_not_in_lookup=True,
            )
            complete_query = complete_query.where(Criterion.all(having))

        if self.order_by:
            for order_clause in self.order_by:
                field = self.project.get_field(order_clause["field"])
                order_by_alias = field.alias(with_view=True)
                if order_by_alias in self.cte_alias_lookup:
                    order_by_alias = f"{self.cte_alias_lookup[order_by_alias]}.{order_by_alias}"
                else:
                    self._raise_query_error_from_cte(field.id())

                order = Order.desc if order_clause.get("sort", "asc").lower() == "desc" else Order.asc
                complete_query = complete_query.orderby(
                    LiteralValue(order_by_alias), order=order, nulls=NullSorting.last
                )

        sql = str(complete_query.limit(self.limit))
        if semicolon:
            sql += ";"
        return sql

    def build_cte_from(self):
        base_cte_query = self._base_query()
        for query in self.merged_queries:
            base_cte_query = base_cte_query.with_(query["query"], query["cte_alias"])

        base_cte_alias = self.merged_queries[0]["cte_alias"]
        base_cte_query = base_cte_query.from_(AliasedQuery(base_cte_alias))

        # We're starting on the second one because the first one is the base in the from statement,
        # and all other queries are joined to it
        for query in self.merged_queries[1:]:
            # We have to do this because Redshift doesn't support a full outer join
            # of two CTE's without dimensions using 1=1
            if self.query_type == Definitions.redshift and len(query["join_fields"]) == 0:
                base_cte_query = base_cte_query.join(AliasedQuery(query["cte_alias"])).cross()
            else:
                criteria = self._build_join_criteria(query["join_fields"], base_cte_alias, query["cte_alias"])
                zenlytic_join_type = query.get("join_type")
                if zenlytic_join_type is None:
                    zenlytic_join_type = ZenlyticJoinType.left_outer
                join_type = self.pypika_join_type_lookup(zenlytic_join_type)
                base_cte_query = base_cte_query.join(AliasedQuery(query["cte_alias"]), how=join_type).on(
                    criteria
                )

        return base_cte_query

    def _build_join_criteria(self, join_logic: list, base_query_alias: str, joined_query_alias: str):
        # Join logic is a list with {'field': field_in_current_cte, 'source_field': field_in_base_cte}
        # No dimensions to join on, the query results must be just one number each
        if len(join_logic) == 0:
            return LiteralValueCriterion("1=1")

        join_criteria = []
        for logic in join_logic:
            joined_field = self.project.get_field(logic["field"])
            base_field = self.project.get_field(logic["source_field"])
            base_alias_and_id = f"{base_query_alias}.{base_field.alias(with_view=True)}"
            joined_alias_and_id = f"{joined_query_alias}.{joined_field.alias(with_view=True)}"
            # We need to add casting for differing datatypes on dimension groups for BigQuery
            if Definitions.bigquery == self.query_type and base_field.datatype != joined_field.datatype:
                join_condition = (
                    f"CAST({base_alias_and_id} AS TIMESTAMP)=CAST({joined_alias_and_id} AS TIMESTAMP)"
                )
            else:
                join_condition = f"{base_alias_and_id}={joined_alias_and_id}"
            join_criteria.append(join_condition)

        return LiteralValueCriterion(" and ".join(join_criteria))

    # Code to handle SELECT portion of query
    def get_select_columns(self):
        self.cte_alias_lookup = {}
        select = []
        existing_aliases = []
        for query in self.merged_queries:
            # We do not want to include the join fields in the SELECT clause,
            # unless they are part of the primary (base) query
            for j in query.get("join_fields", []):
                field = self.project.get_field(j["field"])

                alias = field.alias(with_view=True)
                existing_aliases.append(alias)

            field_ids = query.get("metrics", []) + query.get("dimensions", [])
            for field_id in field_ids:
                field = self.project.get_field(field_id)
                alias = field.alias(with_view=True)

                if alias not in existing_aliases:
                    self.cte_alias_lookup[alias] = query["cte_alias"]
                    select.append(self.sql(f"{query['cte_alias']}.{alias}", alias=alias))
                    existing_aliases.append(alias)

        return select



import os
import re
from glob import glob

import ruamel.yaml
import sqlglot

from .metricflow_types import MetricflowMetricTypes


class ZenlyticUnsupportedError(Exception):
    pass


def convert_mf_project_to_zenlytic_project(
    mf_project: dict,
    project_name: str = "mf_project_name",
    connection_name: str = "mf_connection_name",
    model_dict: dict = {},
) -> tuple:
    """mf_project is a dict with keys for each semantic model
    and the dims, measures, and metrics associated with it
    """
    all_measures = []
    for semantic_model in mf_project.values():
        all_measures.extend(semantic_model.get("measures", []))

    primary_key_mapping = get_primary_key_mapping(mf_project)
    model = {**model_dict, "version": 1, "type": "model", "name": project_name, "connection": connection_name}
    views, errors = [], []
    for _, semantic_model in mf_project.items():
        view, view_errors = convert_mf_view_to_zenlytic_view(
            semantic_model, model["name"], all_measures, primary_key_mapping
        )
        views.append(view)
        errors.extend(view_errors)

    return [model], views, errors


def get_primary_key_mapping(semantic_models: dict) -> dict:
    """
    Extract primary key identifiers from semantic models and create a mapping
    from primary key to view name.

    Args:
        semantic_models: Dictionary of semantic models with their configurations

    Returns:
        Dictionary mapping primary key names to view names
    """
    pk_mapping = {}

    for view_name, semantic_model in semantic_models.items():
        entities = semantic_model.get("entities", [])

        # Find the primary entity (primary key)
        for entity in entities:
            if entity.get("type") == "primary":
                pk_name = entity.get("name")
                if pk_name:
                    pk_mapping[pk_name] = view_name
                break

    return pk_mapping


def load_mf_project(models_folder: str):
    semantic_models, metrics = {}, []
    for fn in read_mf_project_files(models_folder):
        mf_model_dict = convert_yml_to_dict(fn)

        metrics.extend(mf_model_dict.get("metrics", []))
        for semantic_model in mf_model_dict.get("semantic_models", []):
            semantic_models[semantic_model["name"]] = semantic_model
            # Empty list of metrics to be filled below
            semantic_models[semantic_model["name"]]["metrics"] = []

    # Assign metrics to the view they should logically live in
    for metric in metrics:
        type_params = metric["type_params"]
        referenced_metrics = []

        if metric["type"] in {MetricflowMetricTypes.simple, MetricflowMetricTypes.cumulative}:
            referenced_metrics.append(get_name_or_string_literal(type_params["measure"]))
        elif metric["type"] == MetricflowMetricTypes.ratio:
            referenced_metrics.append(get_name_or_string_literal(type_params["numerator"]))
            referenced_metrics.append(get_name_or_string_literal(type_params["denominator"]))
        elif metric["type"] == MetricflowMetricTypes.derived:
            for ref_metric in type_params["metrics"]:
                referenced_metrics.append(ref_metric["name"])

        # First try to find a semantic model that has a measure matching any referenced metric/measure
        assigned = False
        for ref_item in referenced_metrics:
            if assigned:
                break

            for model_name, semantic_model in semantic_models.items():
                if assigned:
                    break

                # Check if the reference is a measure in this semantic model
                for measure in semantic_model.get("measures", []):
                    if ref_item == measure["name"]:
                        semantic_models[model_name]["metrics"].append(metric)
                        assigned = True
                        break

        # If not assigned yet, check if any referenced item is another metric already assigned to a model
        if not assigned:
            for ref_item in referenced_metrics:
                if assigned:
                    break

                for model_name, semantic_model in semantic_models.items():
                    if assigned:
                        break

                    # Check if the reference is a metric already assigned to this semantic model
                    for existing_metric in semantic_model.get("metrics", []):
                        if ref_item == existing_metric["name"]:
                            semantic_models[model_name]["metrics"].append(metric)
                            assigned = True
                            break

        if not assigned:
            print(f"WARNING: Could not assign metric {metric['name']} to any semantic model")

    return semantic_models


def convert_mf_view_to_zenlytic_view(
    mf_semantic_model: dict,
    model_name: str,
    all_measures: list,
    primary_key_mapping: dict,
    original_file_path: str = None,
) -> tuple:
    def _error_func(error, extra: dict = {}):
        return {**extra, "view_name": mf_semantic_model["name"], "message": error}

    errors = []
    zenlytic_data = {"version": 1, "type": "view", "model_name": model_name, "fields": [], "identifiers": []}

    mf_metrics = mf_semantic_model.get("metrics", [])

    if original_file_path:
        zenlytic_data["original_file_path"] = original_file_path

    # Get view-level values
    zenlytic_data["name"] = mf_semantic_model["name"]
    if "sql_table_name" in mf_semantic_model.get("meta", {}):
        zenlytic_data["sql_table_name"] = mf_semantic_model["meta"]["sql_table_name"]
    else:
        zenlytic_data["sql_table_name"] = extract_inner_text(mf_semantic_model["model"])

    if mf_semantic_model.get("config", {}).get("meta") and isinstance(
        mf_semantic_model["config"]["meta"], dict
    ):
        zenlytic_data = {**zenlytic_data, **mf_semantic_model["config"]["meta"].get("zenlytic", {})}

    if description := mf_semantic_model.get("description"):
        zenlytic_data["description"] = description
    default_date = mf_semantic_model.get("defaults", {}).get("agg_time_dimension")

    if default_date:
        zenlytic_data["default_date"] = default_date

    # dimensions to fields.dimensions
    for dimension in mf_semantic_model.get("dimensions", []):
        try:
            field_dict = convert_mf_dimension_to_zenlytic_dimension(dimension)
            zenlytic_data["fields"].append(field_dict)
        except ZenlyticUnsupportedError as e:
            errors.append(_error_func(f"In view {mf_semantic_model['name']} {str(e)}"))

    # measures to measures
    for measure in mf_semantic_model.get("measures", []):
        try:
            field_dict = convert_mf_measure_to_zenlytic_measure(measure)
            zenlytic_data["fields"].append(field_dict)
        except ZenlyticUnsupportedError as e:
            errors.append(_error_func(f"In view {mf_semantic_model['name']} {str(e)}"))

    for metric in mf_metrics:
        try:
            metric_dict, added_measures = convert_mf_metric_to_zenlytic_measure(
                metric, all_measures, primary_key_mapping
            )
            zenlytic_data["fields"].append(metric_dict)
            zenlytic_data["fields"].extend(added_measures)
        except ZenlyticUnsupportedError as e:
            errors.append(_error_func(f"In view {mf_semantic_model['name']} {str(e)}"))

    # entities to identifiers
    for entity in mf_semantic_model["entities"]:
        if "name" in entity:
            identifier = convert_mf_entity_to_zenlytic_identifier(entity, fields=zenlytic_data["fields"])
            zenlytic_data["identifiers"].append(identifier)

    return zenlytic_data, errors


def convert_mf_dimension_to_zenlytic_dimension(mf_dimension: dict):
    if "expr" in mf_dimension:
        sql = append_table_reference(mf_dimension["expr"])
    else:
        sql = "${TABLE}." + mf_dimension["name"]

    field_dict = {"name": mf_dimension["name"], "sql": sql}

    if mf_dimension["type"].lower() == "time":
        field_dict["field_type"] = "dimension_group"
        field_dict["type"] = "time"
        field_dict["timeframes"] = ["raw", "date", "week", "month", "quarter", "year", "month_of_year"]

    elif mf_dimension["type"].lower() == "categorical":
        field_dict["field_type"] = "dimension"
        field_dict["type"] = "string"
        field_dict["searchable"] = True
    else:
        raise ZenlyticUnsupportedError(
            f"field conversion failed for {mf_dimension['name']}: Dimension type {mf_dimension['type']} not"
            " supported"
        )

    if description := mf_dimension.get("description"):
        field_dict["description"] = description

    if label := mf_dimension.get("label"):
        field_dict["label"] = label

    if mf_dimension.get("meta") and isinstance(mf_dimension["meta"], dict):
        field_dict = {**field_dict, **mf_dimension["meta"].get("zenlytic", {})}

    if mf_dimension.get("config", {}).get("meta") and isinstance(mf_dimension["config"]["meta"], dict):
        field_dict = {**field_dict, **mf_dimension["config"]["meta"].get("zenlytic", {})}

    return field_dict


def convert_mf_measure_to_zenlytic_measure(mf_measure: dict):
    field_dict = {"name": mf_measure["name"], "type": mf_measure["agg"], "field_type": "measure"}
    if "expr" in mf_measure:
        field_dict["sql"] = append_table_reference(str(mf_measure["expr"]))
    else:
        field_dict["sql"] = "${TABLE}." + mf_measure["name"]

    if not mf_measure.get("create_metric", False):
        field_dict["hidden"] = True
        # If we are not creating this metric directly, we need to set an underscore
        # in front of the metric name to stop collisions with metrics that have
        # the exact same name
        field_dict["name"] = "_" + field_dict["name"]

    if field_dict["type"] == "sum_boolean":
        field_dict["type"] = "sum"
        field_dict["sql"] = f"CAST({field_dict['sql']} AS INT)"

    if field_dict["type"] == "percentile":
        field_dict["type"] = "percentile"
        field_dict["percentile"] = int(float(mf_measure["agg_params"]["percentile"]) * 100)
        if mf_measure["agg_params"]["use_discrete_percentile"]:
            raise ZenlyticUnsupportedError(
                f"discrete percentile is not supported for the measure {mf_measure['name']}"
            )

    if canon_date := mf_measure.get("agg_time_dimension"):
        field_dict["canon_date"] = canon_date

    if description := mf_measure.get("description"):
        field_dict["description"] = description

    if label := mf_measure.get("label"):
        field_dict["label"] = label

    if "non_additive_dimension" in mf_measure:
        field_dict["non_additive_dimension"] = {
            **mf_measure["non_additive_dimension"],
            # We need to add the dimension group to the non_additive_dimension reference
            "name": mf_measure["non_additive_dimension"]["name"] + "_raw",
        }

    if mf_measure.get("config", {}).get("meta") and isinstance(mf_measure["config"]["meta"], dict):
        field_dict = {**field_dict, **mf_measure["config"]["meta"].get("zenlytic", {})}

    return field_dict


def convert_mf_entity_to_zenlytic_identifier(mf_entity: dict, fields: list = []):
    # if expr is a simple string, use it as the sql otherwise use it as given as a sql snippet
    entity_dict = {
        "name": mf_entity["name"],
        "type": "primary" if mf_entity["type"] in {"unique", "natural", "primary"} else "foreign",
    }
    if "expr" in mf_entity:
        sql_expr = append_table_reference(str(mf_entity["expr"]))
    else:
        sql_expr = "${TABLE}." + mf_entity["name"]
    entity_dict["sql"] = sql_expr

    for f in fields:
        # If the sql expression equals the field name, use a reference to the field
        if "${TABLE}." + f["name"] == sql_expr:
            entity_dict["sql"] = "${" + f["name"] + "}"
            break

    if mf_entity.get("config", {}).get("meta") and isinstance(mf_entity["config"]["meta"], dict):
        entity_dict = {**entity_dict, **mf_entity["config"]["meta"].get("zenlytic", {})}

    return entity_dict


def convert_mf_metric_to_zenlytic_measure(
    mf_metric: dict, measures: list, primary_key_mapping: dict
) -> tuple:
    """This returns a list because metrics with filters applied can
    result in an additional measure(s) being created
    """
    metric_dict = {
        "name": mf_metric["name"],
        "label": mf_metric.get("label", mf_metric["name"].replace("_", " ").title()),
        "field_type": "measure",
    }

    additional_measures = []
    if mf_metric["type"].lower() == MetricflowMetricTypes.cumulative:
        raise ZenlyticUnsupportedError(
            f"metric conversion failed for {mf_metric['name']}: It is a cumulative metric, which is not"
            " supported."
        )

    elif mf_metric["type"].lower() == MetricflowMetricTypes.simple:
        measure_name = get_name_or_string_literal(mf_metric["type_params"]["measure"])
        associated_measure = _get_measure(measure_name, measures, metric_name=mf_metric["name"])
        metric_dict, _ = apply_filter_to_metric(
            associated_measure, mf_metric, primary_key_mapping, extra_metric_params=metric_dict
        )

    elif mf_metric["type"].lower() == MetricflowMetricTypes.ratio:
        metric_dict["type"] = "ratio"
        numerator = mf_metric["type_params"]["numerator"]
        denominator = mf_metric["type_params"]["denominator"]
        numerator_measure_name = get_name_or_string_literal(numerator)
        denominator_measure_name = get_name_or_string_literal(denominator)

        # If there's a filter, re-write the sql to include the filter
        if isinstance(numerator, dict) and "filter" in numerator:
            associated_numerator = _get_measure(numerator["name"], measures, metric_name=mf_metric["name"])
            numerator_dict, numerator_measures = apply_filter_to_metric(
                associated_numerator,
                numerator,
                primary_key_mapping,
                new_measure_name=mf_metric["name"] + "_numerator",
            )
            numerator_sql = "${" + numerator_dict["name"] + "}"
            additional_measures.extend(numerator_measures)
        else:
            numerator_sql = "${" + numerator_measure_name + "}"

        # If there's a filter, re-write the sql to include the filter
        if isinstance(denominator, dict) and "filter" in denominator:
            associated_denominator = _get_measure(
                denominator["name"], measures, metric_name=mf_metric["name"]
            )
            denominator_dict, denominator_measures = apply_filter_to_metric(
                associated_denominator,
                denominator,
                primary_key_mapping,
                new_measure_name=mf_metric["name"] + "_denominator",
            )
            denominator_sql = "${" + denominator_dict["name"] + "}"
            additional_measures.extend(denominator_measures)
        else:
            denominator_sql = "${" + denominator_measure_name + "}"

        metric_dict["sql"] = numerator_sql + " / " + denominator_sql
        metric_dict["type"] = "number"

    elif mf_metric["type"].lower() == MetricflowMetricTypes.derived:
        metric_dict["type"] = "number"
        expr = mf_metric["type_params"]["expr"]
        referenced_metrics = mf_metric["type_params"]["metrics"]

        # Sort metrics by length in descending order to avoid substring replacement issues
        # This ensures longer names are replaced first (e.g., total_gross_revenue_from_advertising before total_gross_revenue)
        for metric in sorted(referenced_metrics, key=lambda x: len(x.get("name", "")), reverse=True):
            if "alias" in metric and "filter" not in metric:
                # Use word boundaries to ensure we only replace whole words/identifiers
                pattern = r"\b" + re.escape(metric["alias"]) + r"\b"
                expr = re.sub(pattern, "${" + metric["name"] + "}", expr)
            elif "alias" in metric and "filter" in metric:
                associated_measure = _get_measure(metric["name"], measures, metric_name=mf_metric["name"])
                measure_dict, added_measures = apply_filter_to_metric(
                    associated_measure,
                    metric,
                    primary_key_mapping,
                    new_measure_name=mf_metric["name"] + f"_{metric['alias']}",
                )
                additional_measures.extend(added_measures)
                pattern = r"\b" + re.escape(metric["alias"]) + r"\b"
                expr = re.sub(pattern, "${" + measure_dict["name"] + "}", expr)
            else:
                # If there is no alias and no filters we just need to add reference syntax
                pattern = r"\b" + re.escape(metric["name"]) + r"\b"
                expr = re.sub(pattern, "${" + metric["name"] + "}", expr)
        metric_dict["sql"] = expr

    else:
        raise ZenlyticUnsupportedError(f"Metric type {mf_metric['type']} not supported")

    if description := mf_metric.get("description"):
        metric_dict["description"] = description

    if "agg_time_dimension" in mf_metric:
        metric_dict["canon_date"] = mf_metric["agg_time_dimension"]

    if mf_metric.get("config", {}).get("meta") and isinstance(mf_metric["config"]["meta"], dict):
        metric_dict = {**metric_dict, **mf_metric["config"]["meta"].get("zenlytic", {})}

    return metric_dict, additional_measures


def _get_measure(measure_name: str, measures: list, metric_name: str):
    try:
        return next((m for m in measures if m["name"] == measure_name))
    except StopIteration:
        raise ZenlyticUnsupportedError(
            f"could not find associated measure {measure_name} referenced in metric {metric_name}"
        )


def apply_filter_to_metric(
    mf_measure: dict,
    mf_metric: dict,
    primary_key_mapping: dict,
    extra_metric_params: dict = {},
    new_measure_name: str = None,
):
    measure_dict = convert_mf_measure_to_zenlytic_measure(mf_measure)
    hidden = not mf_metric.get("config", {}).get("enabled", True)
    metric_dict = {**measure_dict, **extra_metric_params, "hidden": hidden}

    # If there's a filter, re-write the sql to include the filter
    additional_measures = []
    if "filter" in mf_metric:
        try:
            metric_dict["sql"] = apply_filter_to_sql(
                metric_dict["sql"], mf_metric["filter"], primary_key_mapping
            )
        except ValueError as e:
            raise ZenlyticUnsupportedError(f"metric conversion failed for {mf_metric['name']}: {str(e)}")
        if new_measure_name:
            metric_dict["name"] = new_measure_name
            additional_measures.append(metric_dict)
    return metric_dict, additional_measures


def apply_filter_to_sql(sql, filter, primary_key_mapping: dict):
    filter_sql = _extract_filter_sql(filter, primary_key_mapping)
    return f"case when {filter_sql} then {sql} else null end"


def _extract_filter_sql(filter_string, primary_key_mapping: dict):
    """A filter will look like
    "{{ Dimension('order__is_food_order') }} = True
    We want to turn it into a valid filter statement like ${order.is_food_order} = True
    We do NOT currently support Metric or Entity type filters
    """
    if "Entity(" in filter_string:
        raise ValueError("Entity type filters are not supported")
    if "Metric(" in filter_string:
        raise ValueError("Metric type filters are not supported")
    matches = re.findall(r"{{\s*Dimension\('([^']+)'\)\s*}}", filter_string)
    for match in matches:
        column_name = match.replace("__", ".")

        if "." in column_name:
            primary_key_reference, field_name = column_name.split(".")
            if primary_key_reference in primary_key_mapping:
                column_name = primary_key_mapping[primary_key_reference] + "." + field_name

        replacement = "${" + column_name + "}"
        filter_string = filter_string.replace(f"Dimension('{match}')", replacement)

    # First get the dimension name and grain
    time_dim_matches = re.findall(
        r"""TimeDimension\(\s*['"]([^'"]+)['"]\s*,\s*['"]([^'"]+)['"]\s*\)""", filter_string
    )
    for match in time_dim_matches:
        column_name = match[0].replace("__", ".")
        if "." in column_name:
            primary_key_reference, field_name = column_name.split(".")
            if primary_key_reference in primary_key_mapping:
                column_name = primary_key_mapping[primary_key_reference] + "." + field_name

        time_grain = match[1].replace("day", "date")

        replacement = "${" + f"{column_name}_{time_grain}" + "}"
        filter_string = filter_string.replace(f"TimeDimension('{match[0]}', '{match[1]}')", replacement)
    return filter_string.replace("{{", "").replace("}}", "")


def append_table_reference(sql: str):
    try:
        # Parse the SQL to identify column references
        parsed = sqlglot.parse_one(sql.strip())

        # Transform column references to include ${TABLE}. prefix
        def transform_columns(node):
            if isinstance(node, sqlglot.expressions.Column) and not node.table:
                # Only add ${TABLE}. if the column doesn't already have a table reference
                node.set("table", sqlglot.expressions.Identifier(this="${TABLE}"))
            return node

        # Apply the transformation
        transformed = parsed.transform(transform_columns)
        return str(transformed)
    except Exception:
        # Fallback to simple string concatenation if parsing fails
        return sql.strip()


def get_name_or_string_literal(s):
    if isinstance(s, str):
        return s
    elif isinstance(s, dict):
        return s["name"]
    else:
        raise ValueError(f"Invalid type: {type(s)}")


def sql_has_operations(sql: str) -> bool:
    return any(op in sql for op in ["+", "-", "*", "/", "(", ")", " ", ","])


def convert_yml_to_dict(path):
    yaml = ruamel.yaml.YAML(typ="rt")
    yaml.version = (1, 1)
    with open(path, "r") as f:
        yaml_dict = yaml.load(f)
    return yaml_dict


def extract_inner_text(s):
    match = re.search(r"ref\('(.*)'\)", s)
    if match:
        return match.group(1)
    return None


def zenlytic_views_to_yaml(zenlytic_models, zenlytic_views, directory: str = None, write_to_file=True):
    view_directory = os.path.join(directory, "views") if directory else "./views"
    model_directory = os.path.join(directory, "models") if directory else "./models"

    if not os.path.exists(view_directory) and write_to_file:
        os.makedirs(view_directory)

    if not os.path.exists(model_directory) and write_to_file:
        os.makedirs(model_directory)

    zenlytic_yaml = []
    for zenlytic_file in zenlytic_models + zenlytic_views:
        # write the yaml to views/model_name.yml
        if write_to_file:
            if "original_file_path" in zenlytic_file:
                file_path = zenlytic_file["original_file_path"]
            else:
                file_path = f"{zenlytic_file['name']}_{zenlytic_file['type']}.yml"

            if zenlytic_file["type"] == "model":
                write_to_path = os.path.join(model_directory, file_path)
            else:
                write_to_path = os.path.join(view_directory, file_path)
            # write the yaml to views/model_name.yml
            dump_yaml_to_file(zenlytic_file, write_to_path)

        # add the yaml string to views_yaml
        zenlytic_yaml.append(dump_yaml_to_file(zenlytic_file))

    return zenlytic_yaml


def dump_yaml_to_file(data, path: str = None):
    filtered_data = {k: v for k, v in data.items() if not k.startswith("_")}
    if path is None:
        return ruamel.yaml.dump(filtered_data, Dumper=ruamel.yaml.RoundTripDumper)
    else:
        with open(path, "w") as f:
            ruamel.yaml.dump(filtered_data, f, Dumper=ruamel.yaml.RoundTripDumper)


def read_mf_project_files(models_folder: str):
    """Returns a list of all the yml files in the Metricflow project.
    Args:
        models_folder (str): The path to the models folder (usually project_name/models)
    """
    # Make sure models_folder ends with a slash to properly join paths
    if not models_folder.endswith("/"):
        models_folder = models_folder + "/"

    yml_files = glob(f"{models_folder}**/*.yml", recursive=True)
    yaml_files = glob(f"{models_folder}**/*.yaml", recursive=True)
    return yml_files + yaml_files
